@@ -1,5 +1,7 @@
 from . import pdfrenderer
 from . import mrc
-from . import scandata
-from . import pagenumbers
 from . import recode
+from . import scandata
+from . import pdfhacks
+from . import pagenumbers
+from . import grayconvert
