@@ -1,3 +1,19 @@
+# archive-pdf-tools
+# Copyright (C) 2020-2021, Internet Archive
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Affero General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Affero General Public License for more details.
+#
+# You should have received a copy of the GNU Affero General Public License
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+#
 # Author: Merlijn Boris Wolf Wajer <merlijn@archive.org>
 
 import sys
@@ -17,7 +33,7 @@ from skimage.restoration import denoise_tv_bregman, estimate_sigma
 from scipy import ndimage
 import numpy as np
 
-from optimiser import optimise_gray, optimise_rgb, optimise_gray2, optimise_rgb2
+from optimiser import optimise_gray, optimise_rgb, optimise_gray2, optimise_rgb2, fast_mask_denoise
 from sauvola import binarise_sauvola
 
 import fitz
@@ -26,7 +42,8 @@ fitz.TOOLS.set_icc(True) # For good measure, not required
 
 from internetarchivepdf.const import (RECODE_RUNTIME_WARNING_TOO_SMALL_TO_DOWNSAMPLE, JPEG2000_IMPL_KAKADU,
         JPEG2000_IMPL_OPENJPEG, JPEG2000_IMPL_GROK, JPEG2000_IMPL_PILLOW,
-        COMPRESSOR_JPEG, COMPRESSOR_JPEG2000)
+        COMPRESSOR_JPEG, COMPRESSOR_JPEG2000, DENOISE_NONE, DENOISE_FAST,
+        DENOISE_BREGMAN)
 
 
 """
@@ -47,11 +64,11 @@ def mean_estimate_sigma(arr):
         return np.mean(estimate_sigma(arr))
 
 
-def threshold_image(img, dpi):
+def threshold_image(img, dpi, k=0.34):
     window_size = 51
 
     if dpi is not None:
-        window_size = int(dpi / 2)
+        window_size = int(dpi / 4)
         if window_size % 2 == 0:
             window_size += 1
 
@@ -60,7 +77,7 @@ def threshold_image(img, dpi):
     out_img = np.reshape(out_img, w*h)
     in_img = np.reshape(img, w*h)
 
-    binarise_sauvola(in_img, out_img, w, h, window_size, window_size, 0.34, 128)
+    binarise_sauvola(in_img, out_img, w, h, window_size, window_size, k, 128)
     out_img = np.reshape(out_img, (h, w))
     # TODO: optimise this, we can do it in binarise_sauvola
     out_img = np.invert(out_img)
@@ -174,10 +191,6 @@ def create_hocr_mask(img, mask_arr, hocr_word_data, downsample=None, dpi=None, t
             if line_text.strip() == '' or line_conf < 20:
                 continue
 
-
-            # TODO: maybe work with font size per line and word confidence in
-            # the line
-
             if downsample is not None:
                 coords = [int(x/downsample) for x in coords]
             else:
@@ -200,37 +213,103 @@ def create_hocr_mask(img, mask_arr, hocr_word_data, downsample=None, dpi=None, t
             # Simple grayscale invert
             np_lineimg_invert = 255 - np.copy(np_lineimg)
 
-            thres = threshold_image(np_lineimg, dpi)
+            # XXX: If you tweak k, you must tweak the various ratio and sigma's
+            # based on the test images
+            k = 0.1
+            thres = threshold_image(np_lineimg, dpi, k)
             ones = np.count_nonzero(thres)
             zero = (img.size[0] * img.size[1]) - ones
             ratio = (ones/(zero+ones))*100
 
-            thres_invert = threshold_image(np_lineimg_invert, dpi)
-            ones = np.count_nonzero(thres_invert)
-            zero = (img.size[0] * img.size[1]) - ones
-            inv_ratio = (ones/(zero+ones))*100
+            thres_invert = threshold_image(np_lineimg_invert, dpi, k)
+            ones_i = np.count_nonzero(thres_invert)
+            zero_i = (img.size[0] * img.size[1]) - ones
+            inv_ratio = (ones_i/(zero_i+ones_i))*100
 
-            if ratio < 0.2 or inv_ratio < 0.2:
+            if ratio < 0.3 or inv_ratio < 0.3:
                 th = None
+
                 perc_larger = 0.
                 if inv_ratio != 0.0:
                     perc_larger = (ratio / inv_ratio) * 100
 
-                # Prefer ratio over inv_ratio by a bit
-                if (ratio < inv_ratio or perc_larger < 110.) and ratio < 0.2:
+                if inv_ratio > 0.2 and ratio < 0.2:
                     th = thres
-                elif inv_ratio < 0.2:
-                    th = thres_invert
+                else:
+                    # mean_estimate_sigma is expensive, so let's only do it if
+                    # we need to
+
+                    ratio_sigma = mean_estimate_sigma(thres)
+                    inv_ratio_sigma = mean_estimate_sigma(thres_invert)
+
+
+                    # Prefer ratio over inv_ratio by a bit
+                    if inv_ratio < 0.3 and inv_ratio < ratio and \
+                    (inv_ratio_sigma < ratio_sigma or \
+                    (ratio_sigma < 0.1 and inv_ratio_sigma < 0.1)):
+                        th = thres_invert
+                    elif ratio < 0.2:
+                        th = thres
 
                 if th is not None:
                     mask_arr[top:bottom, left:right] = th
 
+
     if timing_data is not None:
         timing_data.append(('hocr_mask_gen', time() - t))
 
 
+def estimate_noise(imgf):
+    #sigma_est = mean_estimate_sigma(imgf)
+    #return sigma_est
+
+    # We do this only on a part of the image, because it's accurate enough wrt
+    # noise estimation (definitely for camera noise estimation since that's
+    # everywhere in the image, and it's quite a bit faster this way).
+    h, w = imgf.shape
+    MUL = 4
+    hs = int(h/2 - h/MUL)
+    he = int(h/2 + h/MUL)
+    ws = int(w/2 - w/MUL)
+    we = int(w/2 + w/MUL)
+
+    # Really small image?
+    if he == 0 or we == 0:
+        hs = 0
+        he = h
+        ws = 0
+        we = w
+
+    sigma_est = mean_estimate_sigma(imgf[hs:he, ws:we])
+
+    return sigma_est
+
+
 
 def create_threshold_mask(mask_arr, imgf, dpi=None, denoise_mask=None, timing_data=None):
+    # We don't apply any of these blurs to the hOCR mask, we want that as
+    # sharp as possible.
+
+    t = time()
+    sigma_est = estimate_noise(imgf)
+
+    if timing_data is not None:
+        timing_data.append(('est_1', time() - t))
+    if sigma_est > 1.0:
+        t = time()
+        imgf = ndimage.filters.gaussian_filter(imgf, sigma=sigma_est*0.1)
+        if timing_data is not None:
+            timing_data.append(('blur_1', time() - t))
+
+        #t = time()
+        #n_sigma_est = mean_estimate_sigma(imgf)
+        #time_data.append(('est_2', time() - t))
+        #if sigma_est > 1.0 and n_sigma_est > 1.0:
+            #    t = time()
+        #    imgf = ndimage.filters.gaussian_filter(imgf, sigma=sigma_est*0.5)
+        #    print('Going for second blur: n_sigma_est:',n_sigma_est)
+        #    time_data.append(('blur_2', time() - t))
+
     t = time()
     thres_arr = threshold_image(imgf.astype(np.uint8), dpi)
     if timing_data is not None:
@@ -245,6 +324,7 @@ def create_mrc_hocr_components(image, hocr_word_data,
                                dpi=None,
                                downsample=None,
                                bg_downsample=None,
+                               fg_downsample=None,
                                denoise_mask=None, timing_data=None,
                                errors=None):
     """
@@ -271,6 +351,8 @@ def create_mrc_hocr_components(image, hocr_word_data,
         if timing_data is not None:
             timing_data.append(('grey_conversion', time() - t))
 
+    width_, height_ = image.size
+
     mask_arr = np.array(Image.new('1', image.size))
 
     # Modifies mask_arr in place
@@ -288,11 +370,20 @@ def create_mrc_hocr_components(image, hocr_word_data,
                               denoise_mask=denoise_mask,
                               timing_data=timing_data)
 
-    if denoise_mask:
+    if denoise_mask != DENOISE_NONE:
         t = time()
-        mask_arr = denoise_bregman(mask_arr)
-        if timing_data is not None:
-            timing_data.append(('denoise', time() - t))
+        if denoise_mask == DENOISE_FAST:
+            # XXX: We could make the mincnt parameter take the dpi into account
+            fast_mask_denoise(mask_arr, width_, height_, 4, 2)
+            if timing_data is not None:
+                timing_data.append(('fast_denoise', time() - t))
+        elif denoise_mask == DENOISE_BREGMAN:
+            mask_arr = denoise_bregman(mask_arr)
+            if timing_data is not None:
+                timing_data.append(('denoise', time() - t))
+        else:
+            raise ValueError('Invalid denoise option:', denoise_mask)
+
 
     yield mask_arr
 
@@ -302,7 +393,6 @@ def create_mrc_hocr_components(image, hocr_word_data,
     # Take foreground pixels and optimise the image by making the surrounding
     # pixels like the foreground, allowing for more optimal compression (and
     # higher quality foreground pixels as a result)
-    width_, height_ = image.size
     if image.mode == 'L':
         foreground_arr = optimise_gray2(mask_arr, image_arr, width_, height_, 3)
     else:
@@ -310,6 +400,23 @@ def create_mrc_hocr_components(image, hocr_word_data,
     if timing_data is not None:
         # The name fg_partial_blur is kept for backwards compatibility
         timing_data.append(('fg_partial_blur', time() - t))
+
+    if fg_downsample is not None:
+        t = time()
+        image2 = Image.fromarray(foreground_arr)
+        w, h = image2.size
+        w_downsample = int(w / fg_downsample)
+        h_downsample = int(h / fg_downsample)
+        if w_downsample > 0 and h_downsample > 0:
+            image2.thumbnail((w_downsample, h_downsample))
+            foreground_arr = np.array(image2)
+        else:
+            if errors is not None:
+                errors.add(RECODE_RUNTIME_WARNING_TOO_SMALL_TO_DOWNSAMPLE)
+
+        if timing_data is not None:
+            timing_data.append(('fg_downsample', time() - t))
+
     yield foreground_arr
     foreground_arr = None
 
@@ -348,7 +455,8 @@ def create_mrc_hocr_components(image, hocr_word_data,
     return
 
 
-def encode_mrc_mask(np_mask, tmp_dir=None, jbig2=True, timing_data=None):
+def encode_mrc_mask(np_mask, tmp_dir=None, jbig2=True, embedded_jbig2=False,
+                    timing_data=None):
     """
     Encode mask image either to JBIG2 or PNG.
 
@@ -357,6 +465,7 @@ def encode_mrc_mask(np_mask, tmp_dir=None, jbig2=True, timing_data=None):
     * np_mask (numpy.array): Mask image array
     * tmp_dir (str): path the temporary directory to write images to
     * jbig2 (bool): Whether to encode to JBIG2 or PNG
+    * embedded_jbig2 (bool): Whether to encode to JBIG2 with or without header
     * timing_data (optional): Add time information to timing_data structure
 
     Returns a tuple: (str, str) where the first entry is the jbig2
@@ -374,7 +483,10 @@ def encode_mrc_mask(np_mask, tmp_dir=None, jbig2=True, timing_data=None):
     mask.save(mask_img_png, compress_level=0)
 
     if jbig2:
-        out = subprocess.check_output(['jbig2', mask_img_png])
+        if embedded_jbig2:
+            out = subprocess.check_output(['jbig2', '-p', mask_img_png])
+        else:
+            out = subprocess.check_output(['jbig2', mask_img_png])
         fp= open(mask_img_jbig2, 'wb+')
         fp.write(out)
         fp.close()
@@ -388,6 +500,96 @@ def encode_mrc_mask(np_mask, tmp_dir=None, jbig2=True, timing_data=None):
         return None, mask_img_png
 
 
+def _jpeg2000_pillow_str_to_kwargs(s):
+    import ast
+    kwargs = {}
+    for en in s.split(';'):
+        k, v = en.split(':', maxsplit=1)
+        kwargs[k] = ast.literal_eval(v)
+
+    return kwargs
+
+
+def encode_mrc_img(np_img, img_compression_flags, imgtype=None, tmp_dir=None,
+        jpeg2000_implementation=None, mrc_image_format=None, timing_data=None):
+    """
+    Encode image as JPEG2000 or JPEG, with the provided compression settings
+    and JPEG2000/JPEG encoder.
+
+    Args:
+
+    * np_img (numpy.array): Image array
+    * img_compression_flags (str): Compression flags
+    * imgtype (str: 'bg' or 'fg'
+    * tmp_dir (str): path the temporary directory to write images to
+    * jpeg2000_implementation (str): What JPEG2000 implementation to use
+    * mrc_image_format (str): What image format to produce
+    * timing_data (optional): Add time information to timing_data structure
+
+    Returns the filepath to the JPEG2000 image
+    """
+    t = time()
+    if imgtype not in ('bg', 'fg'):
+        raise ValueError('imgtype should be \'bg\' or \'fg\'')
+
+    # Create background
+    if mrc_image_format == COMPRESSOR_JPEG:
+        fd, img_tiff = mkstemp(prefix=imgtype, suffix='.jpg', dir=tmp_dir)
+        close(fd)
+    else:
+        if jpeg2000_implementation in (JPEG2000_IMPL_KAKADU, JPEG2000_IMPL_GROK):
+            fd, img_tiff = mkstemp(prefix=imgtype, suffix='.tiff', dir=tmp_dir)
+            close(fd)
+        elif jpeg2000_implementation == JPEG2000_IMPL_OPENJPEG:
+            fd, img_tiff = mkstemp(prefix=imgtype, suffix='.pnm', dir=tmp_dir)
+            close(fd)
+
+    fd, img_jp2 = mkstemp(prefix=imgtype, suffix='.jp2', dir=tmp_dir)
+    close(fd)
+    remove(img_jp2) # XXX: Kakadu doesn't want the file to exist, so what are
+                       # we even doing
+
+    img = Image.fromarray(np_img)
+
+    if mrc_image_format == COMPRESSOR_JPEG:
+        img.save(img_tiff, quality=100)
+
+        output = subprocess.check_output(['jpegoptim'] + img_compression_flags +
+                [img_tiff, '--stdout'])
+        tmpfd=open(img_jp2, 'bw+') # XXX: FIXME: this defeats the point of a tmpfile
+        tmpfd.write(output)
+        tmpfd.close()
+    else:
+        if jpeg2000_implementation == JPEG2000_IMPL_PILLOW:
+            kwargs = _jpeg2000_pillow_str_to_kwargs(img_compression_flags[0])
+            img.save(img_jp2, **kwargs)
+        else:
+            img.save(img_tiff)
+
+            if jpeg2000_implementation == JPEG2000_IMPL_KAKADU:
+                subprocess.check_call([KDU_COMPRESS,
+                    '-num_threads', '0',
+                    '-i', img_tiff, '-o', img_jp2] + img_compression_flags,
+                    stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
+            elif jpeg2000_implementation == JPEG2000_IMPL_OPENJPEG:
+                subprocess.check_call([OPJ_COMPRESS,
+                    '-i', img_tiff, '-o', img_jp2] + img_compression_flags,
+                    stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
+            elif jpeg2000_implementation == JPEG2000_IMPL_GROK:
+                subprocess.check_call([GRK_COMPRESS, '-H', '1',
+                    '-i', img_tiff, '-o', img_jp2] + img_compression_flags,
+                    stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
+            else:
+                raise Exception('Error: invalid jpeg2000 implementation?')
+
+            remove(img_tiff)
+
+    if timing_data is not None:
+        timing_data.append(('%s_jp2' % imgtype, time()-t))
+
+    return img_jp2
+
+
 def encode_mrc_background(np_bg, bg_compression_flags, tmp_dir=None,
         jpeg2000_implementation=None, mrc_image_format=None, timing_data=None):
     """
@@ -405,56 +607,9 @@ def encode_mrc_background(np_bg, bg_compression_flags, tmp_dir=None,
 
     Returns the filepath to the JPEG2000 background image
     """
-    t = time()
-    # Create background
-    if mrc_image_format == COMPRESSOR_JPEG:
-        fd, bg_img_tiff = mkstemp(prefix='bg', suffix='.jpg', dir=tmp_dir)
-    else:
-        if jpeg2000_implementation in (JPEG2000_IMPL_KAKADU, JPEG2000_IMPL_GROK):
-            fd, bg_img_tiff = mkstemp(prefix='bg', suffix='.tiff', dir=tmp_dir)
-        else:
-            fd, bg_img_tiff = mkstemp(prefix='bg', suffix='.pnm', dir=tmp_dir)
-
-    close(fd)
-    fd, bg_img_jp2 = mkstemp(prefix='bg', suffix='.jp2', dir=tmp_dir)
-    close(fd)
-    remove(bg_img_jp2) # XXX: Kakadu doesn't want the file to exist, so what are
-                       # we even doing
-
-    bg_img = Image.fromarray(np_bg)
-    if mrc_image_format == COMPRESSOR_JPEG:
-        bg_img.save(bg_img_tiff, quality=100)
-    else:
-        bg_img.save(bg_img_tiff)
-
-    if mrc_image_format == COMPRESSOR_JPEG:
-        output = subprocess.check_output(['jpegoptim'] + bg_compression_flags +
-                [bg_img_tiff, '--stdout'])
-        tmpfd=open(bg_img_jp2, 'bw+') # XXX: FIXME: this defeats the point of a tmpfile
-        tmpfd.write(output)
-        tmpfd.close()
-    elif jpeg2000_implementation == JPEG2000_IMPL_KAKADU:
-        subprocess.check_call([KDU_COMPRESS,
-            '-num_threads', '0',
-            '-i', bg_img_tiff, '-o', bg_img_jp2] + bg_compression_flags,
-            stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
-    elif jpeg2000_implementation == JPEG2000_IMPL_OPENJPEG:
-        subprocess.check_call([OPJ_COMPRESS,
-            '-i', bg_img_tiff, '-o', bg_img_jp2] + bg_compression_flags,
-            stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
-    elif jpeg2000_implementation == JPEG2000_IMPL_GROK:
-        subprocess.check_call([GRK_COMPRESS, '-H', '1',
-            '-i', bg_img_tiff, '-o', bg_img_jp2] + bg_compression_flags,
-            stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
-    else:
-        raise Exception('Error: invalid jpeg2000 implementation?')
-
-    remove(bg_img_tiff)
-
-    if timing_data is not None:
-        timing_data.append(('bg_jp2', time()-t))
-
-    return bg_img_jp2
+    return encode_mrc_img(np_bg, bg_compression_flags, 'bg', tmp_dir=tmp_dir,
+            jpeg2000_implementation=jpeg2000_implementation,
+            mrc_image_format=mrc_image_format, timing_data=timing_data)
 
 
 def encode_mrc_foreground(np_fg, fg_compression_flags, tmp_dir=None,
@@ -474,62 +629,17 @@ def encode_mrc_foreground(np_fg, fg_compression_flags, tmp_dir=None,
 
     Returns the filepath to the JPEG2000 foreground image
     """
-    t = time()
-    # Create foreground
-    if mrc_image_format == COMPRESSOR_JPEG:
-        fd, fg_img_tiff = mkstemp(prefix='fg', suffix='.jpg', dir=tmp_dir)
-    else:
-        if jpeg2000_implementation in (JPEG2000_IMPL_KAKADU, JPEG2000_IMPL_GROK):
-            fd, fg_img_tiff = mkstemp(prefix='fg', suffix='.tiff', dir=tmp_dir)
-        else:
-            fd, fg_img_tiff = mkstemp(prefix='fg', suffix='.pnm', dir=tmp_dir)
-
-    close(fd)
-    fd, fg_img_jp2 = mkstemp(prefix='fg', suffix='.jp2', dir=tmp_dir)
-    close(fd)
-    remove(fg_img_jp2) # XXX: Kakadu doesn't want the file to exist, so what are
-                       # we even doing
-
-    fg_img = Image.fromarray(np_fg)
-    if mrc_image_format == COMPRESSOR_JPEG:
-        fg_img.save(fg_img_tiff, quality=100)
-    else:
-        fg_img.save(fg_img_tiff)
-
-    if mrc_image_format == COMPRESSOR_JPEG:
-        output = subprocess.check_output(['jpegoptim'] + fg_compression_flags +
-                [fg_img_tiff, '--stdout'])
-        tmpfd=open(fg_img_jp2, 'bw+') # XXX: FIXME: this defeats the point of a tmpfile
-        tmpfd.write(output)
-        tmpfd.close()
-    elif jpeg2000_implementation == JPEG2000_IMPL_KAKADU:
-        subprocess.check_call([KDU_COMPRESS,
-            '-num_threads', '0',
-            '-i', fg_img_tiff, '-o', fg_img_jp2] + fg_compression_flags,
-            stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
-    elif jpeg2000_implementation == JPEG2000_IMPL_OPENJPEG:
-        subprocess.check_call([OPJ_COMPRESS,
-            '-i', fg_img_tiff, '-o', fg_img_jp2] + fg_compression_flags,
-            stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
-    elif jpeg2000_implementation == JPEG2000_IMPL_GROK:
-        subprocess.check_call([GRK_COMPRESS, '-H', '1',
-            '-i', fg_img_tiff, '-o', fg_img_jp2] + fg_compression_flags,
-            stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
-    else:
-        raise Exception('Error: invalid jpeg2000 implementation?')
-
-    remove(fg_img_tiff)
-
-    if timing_data is not None:
-        timing_data.append(('fg_jp2', time()-t))
-
-    return fg_img_jp2
+    return encode_mrc_img(np_fg, fg_compression_flags, 'fg', tmp_dir=tmp_dir,
+            jpeg2000_implementation=jpeg2000_implementation,
+            mrc_image_format=mrc_image_format, timing_data=timing_data)
 
 
 def encode_mrc_images(mrc_gen, bg_compression_flags=None, fg_compression_flags=None,
                       tmp_dir=None, jbig2=True, timing_data=None,
-                      jpeg2000_implementation=None, mrc_image_format=None):
-    mask_img_jbig2, mask_img_png = encode_mrc_mask(next(mrc_gen), tmp_dir=tmp_dir, jbig2=jbig2,
+                      jpeg2000_implementation=None, mrc_image_format=None,
+                      embedded_jbig2=False):
+    mask_img_jbig2, mask_img_png = encode_mrc_mask(next(mrc_gen),
+            tmp_dir=tmp_dir, jbig2=jbig2, embedded_jbig2=embedded_jbig2,
             timing_data=timing_data)
 
     np_fg = next(mrc_gen)
