@@ -1,3 +1,20 @@
+# archive-pdf-tools
+# Copyright (C) 2020-2021, Internet Archive
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Affero General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Affero General Public License for more details.
+#
+# You should have received a copy of the GNU Affero General Public License
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+#
+# Author: Merlijn Boris Wolf Wajer <merlijn@archive.org>
 import pkg_resources
 
 import sys
@@ -11,15 +28,12 @@ from os.path import join
 import shutil
 import json
 from glob import glob
-from math import ceil
-from xml.sax.saxutils import escape as xmlescape
 import re
 import io
 
 
 from PIL import Image
 from PIL import Jpeg2KImagePlugin
-from skimage.color import rgb2hsv
 import numpy as np
 import fitz
 
@@ -29,22 +43,19 @@ from internetarchivepdf.mrc import KDU_EXPAND, OPJ_DECOMPRESS, \
         GRK_DECOMPRESS, \
         create_mrc_hocr_components, \
         encode_mrc_images, encode_mrc_mask
+from internetarchivepdf.grayconvert import special_gray_convert
+from internetarchivepdf.pdfhacks import fast_insert_image, write_pdfa, \
+        write_page_labels, write_basic_ua, write_metadata
 from internetarchivepdf.pdfrenderer import TessPDFRenderer
-from internetarchivepdf.pagenumbers import parse_series, series_to_pdf
 from internetarchivepdf.scandata import scandata_xml_get_skip_pages, \
         scandata_xml_get_page_numbers, scandata_xml_get_dpi_per_page, \
         scandata_xml_get_document_dpi
-from internetarchivepdf.const import (VERSION, PRODUCER,
-        IMAGE_MODE_PASSTHROUGH, IMAGE_MODE_PIXMAP, IMAGE_MODE_MRC,
-        RECODE_RUNTIME_WARNING_INVALID_PAGE_SIZE,
+from internetarchivepdf.const import (IMAGE_MODE_PASSTHROUGH, IMAGE_MODE_PIXMAP,
+        IMAGE_MODE_MRC, RECODE_RUNTIME_WARNING_INVALID_PAGE_SIZE,
         RECODE_RUNTIME_WARNING_INVALID_PAGE_NUMBERS,
-        RECODE_RUNTIME_WARNING_INVALID_JP2_HEADERS,
-        JPEG2000_IMPL_KAKADU,
-        JPEG2000_IMPL_OPENJPEG,
-        JPEG2000_IMPL_GROK,
-        JPEG2000_IMPL_PILLOW,
-        COMPRESSOR_JPEG2000,
-        COMPRESSOR_JPEG)
+        RECODE_RUNTIME_WARNING_INVALID_JP2_HEADERS, JPEG2000_IMPL_KAKADU,
+        JPEG2000_IMPL_OPENJPEG, JPEG2000_IMPL_GROK, JPEG2000_IMPL_PILLOW,
+        COMPRESSOR_JPEG2000, COMPRESSOR_JPEG)
 
 PDFA_MIN_UNITS = 3
 PDFA_MAX_UNITS = 14400
@@ -77,53 +88,6 @@ def guess_dpi(w, h, expected_format=(8.27, 11.69), round_to=[72, 96, 150, 300, 6
     return sorted_diffs[0][0]
 
 
-perc2val = lambda x: (x*255)/100
-
-def level_arr(arr, minv=0, maxv=255):
-    interval = (maxv/255.) - (minv/255.)
-    arr_zero = arr < minv
-    arr_max = arr > maxv
-    arr[::] = ((arr[::] - minv) / interval)
-    arr[arr_zero] = 0
-    arr[arr_max] = 255
-    return arr
-
-
-# Straight forward port of color2Gray.sh script
-# We might be able to do better, but there are only a few users of this script
-# in the archive.org currently, so more time has not been invested in finding
-# alternative or better ways.
-def special_gray_convert(imd):
-    components = ('r', 'g', 'b')
-
-    d = {}
-    for i, k in enumerate(components):
-        for fun in ['min', 'max', 'mean', 'std']:
-            d[k + '_' + fun] = getattr(np, fun)(imd[:,:,i]) / 255.
-
-    bright_adjust = round(d['r_mean'] * d['g_mean'] * d['b_mean'] /
-                    (d['b_max']*(1-d['r_std'])*(1-d['g_std'])*(1-d['b_std'])), 4)
-
-    low_thres = min(int((196 * d['r_min']+14.5)/1), 50)
-
-    high_thres = {
-            'r': min(int((35.66*bright_adjust+48.5)/1), 95),
-            'g': min(int((39.22*bright_adjust+44.5)/1), 95),
-            'b': min(int((45.16*bright_adjust+36.5)/1), 95),
-            }
-
-    new_imd = np.copy(imd)
-    for i, c in enumerate(components):
-        new_imd[:,:,i] = level_arr(new_imd[:,:,i],
-                                   minv=perc2val(low_thres),
-                                   maxv=perc2val(high_thres[c]))
-
-    hsv = rgb2hsv(new_imd)
-    # Calculate the 'L' from 'HSL' as L = S * (1 - V/2)
-    l = hsv[:,:,2] * (1 - (hsv[:,:,1]/2))
-    return np.array(l * 255, dtype=np.uint8)
-
-
 def create_tess_textonly_pdf(hocr_file, save_path, in_pdf=None,
         image_files=None, dpi=None, skip_pages=None, dpi_pages=None,
         reporter=None,
@@ -190,8 +154,11 @@ def create_tess_textonly_pdf(hocr_file, save_path, in_pdf=None,
                         stdout=subprocess.DEVNULL)
 
                     img = Image.open(tiff_in)
+
+                    # img.load() will release the file descriptor
+                    img.load()
                     size = img.size
-                    del img
+                    img = None
                     os.remove(tiff_in)
                 finally:
                     fd.close()
@@ -329,6 +296,7 @@ def insert_images_mrc(to_pdf, hocr_file, from_pdf=None, image_files=None,
         skip_pages=None, img_dir=None, jbig2=False,
         downsample=None,
         bg_downsample=None,
+        fg_downsample=None,
         denoise_mask=None, reporter=None,
         hq_pages=None, hq_bg_compression_flags=None, hq_fg_compression_flags=None,
         verbose=False, tmp_dir=None, report_every=None,
@@ -396,7 +364,7 @@ def insert_images_mrc(to_pdf, hocr_file, from_pdf=None, image_files=None,
             # Do not subtract skipped pages here
             imgfile = image_files[idx+skipped_pages]
 
-            if imgfile.endswith('.jp2') or imgfile.endswith('.jpx'):
+            if (imgfile.endswith('.jp2') or imgfile.endswith('.jpx')) and jpeg2000_implementation != JPEG2000_IMPL_PILLOW:
                 # TODO: If we have pillow jpeg2000 implementation, let's use it
                 # and not do the whole tmpfile dance.
 
@@ -438,15 +406,25 @@ def insert_images_mrc(to_pdf, hocr_file, from_pdf=None, image_files=None,
                             '-i', imgfile, '-o', tiff_in], stderr=subprocess.DEVNULL,
                             stdout=subprocess.DEVNULL)
                 else:
-                    # TODO: could use pillow fallback, but let's do that before
-                    # we make tmpfiles
                     raise Exception('Error: invalid jpeg2000 implementation?')
 
                 image = Image.open(tiff_in)
                 image.load()
+
+                # Windows complains that Pillow doesn't actually close the fd
+                # (not sure why, .load() should close the fd, but it doesn't
+                # seem to, and we can't use .close() since we need the image
+                # later on
+                if sys.platform == 'win32':
+                    new_img = image.copy()
+                    image.close()
+                    image = new_img
+
                 os.remove(tiff_in)
             else:
                 image = Image.open(imgfile)
+                image.load()
+
             if timing_data is not None:
                 timing_data.append(('image_load', time()-t))
 
@@ -492,8 +470,10 @@ def insert_images_mrc(to_pdf, hocr_file, from_pdf=None, image_files=None,
             ww, hh = image.size
             mrc_gen = create_mrc_hocr_components(image, hocr_word_data,
                     dpi=picked_dpi,
-                    downsample=downsample, bg_downsample=None if render_hq else
-                    bg_downsample, denoise_mask=denoise_mask,
+                    downsample=downsample,
+                    bg_downsample=None if render_hq else bg_downsample,
+                    fg_downsample=None if render_hq else fg_downsample,
+                    denoise_mask=denoise_mask,
                     timing_data=timing_data, errors=errors)
             np_mask = next(mrc_gen)
             np_mask = np_mask ^ np.ones(np_mask.shape, dtype=bool)
@@ -517,8 +497,10 @@ def insert_images_mrc(to_pdf, hocr_file, from_pdf=None, image_files=None,
         else:
             mrc_gen = create_mrc_hocr_components(image, hocr_word_data,
                     dpi=picked_dpi,
-                    downsample=downsample, bg_downsample=None if render_hq else
-                    bg_downsample, denoise_mask=denoise_mask,
+                    downsample=downsample,
+                    bg_downsample=None if render_hq else bg_downsample,
+                    fg_downsample=None if render_hq else fg_downsample,
+                    denoise_mask=denoise_mask,
                     timing_data=timing_data, errors=errors)
 
 
@@ -528,30 +510,46 @@ def insert_images_mrc(to_pdf, hocr_file, from_pdf=None, image_files=None,
             # separately from here so that we can free the arrays sooner (and even
             # get the images separately from the create_mrc_hocr_components call)
 
+            fast_insert_image_ok = jbig2 and image.mode in ('L', 'RGB')
+
             mask_f, bg_f, bg_s, fg_f, fg_s = encode_mrc_images(mrc_gen,
                     bg_compression_flags=hq_bg_compression_flags if render_hq else bg_compression_flags,
                     fg_compression_flags=hq_fg_compression_flags if render_hq else fg_compression_flags,
                     tmp_dir=tmp_dir, jbig2=jbig2, timing_data=timing_data,
                     jpeg2000_implementation=jpeg2000_implementation,
-                    mrc_image_format=mrc_image_format)
+                    mrc_image_format=mrc_image_format,
+                    embedded_jbig2=fast_insert_image_ok)
 
             if img_dir is not None:
                 shutil.copy(mask_f, join(img_dir, '%.6d_mask.jbig2' % idx))
                 shutil.copy(bg_f, join(img_dir, '%.6d_bg.jp2' % idx))
                 shutil.copy(fg_f, join(img_dir, '%.6d_fg.jp2' % idx))
 
+
             t = time()
             bg_contents = open(bg_f, 'rb').read()
-            # Tell PyMuPDF about width/height/alpha since it's faster this way
-            page.insert_image(page.rect, stream=bg_contents, mask=None,
+            if not jbig2 or image.mode not in ('L', 'RGB'):
+                # Tell PyMuPDF about width/height/alpha since it's faster this way
+                page.insert_image(page.rect, stream=bg_contents, mask=None,
                     overlay=False, width=bg_s[0], height=bg_s[1], alpha=0)
+            else:
+                fast_insert_image(page, page.rect, stream=bg_contents,
+                                  mask=None, width=bg_s[0], height=bg_s[1],
+                                  stream_fmt=mrc_image_format,
+                                  gray=image.mode == 'L')
 
             fg_contents = open(fg_f, 'rb').read()
             mask_contents = open(mask_f, 'rb').read()
 
             # Tell PyMuPDF about width/height/alpha since it's faster this way
-            page.insert_image(page.rect, stream=fg_contents, mask=mask_contents,
-                    overlay=True, width=fg_s[0], height=fg_s[1], alpha=0)
+            if not jbig2 or image.mode not in ('L', 'RGB'):
+                page.insert_image(page.rect, stream=fg_contents, mask=mask_contents,
+                        overlay=True, width=fg_s[0], height=fg_s[1], alpha=0)
+            else:
+                fast_insert_image(page, page.rect, stream=fg_contents,
+                                  mask=mask_contents, width=fg_s[0], height=fg_s[1],
+                                  stream_fmt=mrc_image_format,
+                                  gray=image.mode == 'L')
 
             # Remove leftover files
             remove(mask_f)
@@ -629,328 +627,6 @@ def insert_images(from_pdf, to_pdf, mode, report_every=None, stop_after=None):
             sys.stdout.flush()
 
 
-# XXX: tmp.icc - pick proper one and ship it with the tool, or embed it
-def write_pdfa(to_pdf):
-    srgbxref = to_pdf.get_new_xref()
-    to_pdf.update_object(srgbxref, """
-<<
-      /Alternate /DeviceRGB
-      /N 3
->>
-""")
-    icc = pkg_resources.resource_string('internetarchivepdf', "data/tmp.icc")
-    to_pdf.update_stream(srgbxref, icc, new=True)
-
-    intentxref = to_pdf.get_new_xref()
-    to_pdf.update_object(intentxref, """
-<<
-  /Type /OutputIntent
-  /S /GTS_PDFA1
-  /OutputConditionIdentifier (Custom)
-  /Info (sRGB IEC61966-2.1)
-  /DestOutputProfile %d 0 R
->>
-""" % srgbxref)
-
-    catalogxref = to_pdf.pdf_catalog()
-    s = to_pdf.xref_object(to_pdf.pdf_catalog())
-    s = s[:-2]
-    s += '  /OutputIntents [ %d 0 R ]' % intentxref
-    s += '>>'
-    to_pdf.update_object(catalogxref, s)
-
-
-def write_page_labels(to_pdf, scandata, errors=None):
-    page_numbers = scandata_xml_get_page_numbers(scandata)
-    res, all_ok = parse_series(page_numbers)
-
-    # Add warning/error
-    if errors is not None and not all_ok:
-        errors.add(RECODE_RUNTIME_WARNING_INVALID_PAGE_NUMBERS)
-
-    catalogxref = to_pdf.pdf_catalog()
-    s = to_pdf.xref_object(to_pdf.pdf_catalog())
-    s = s[:-2]
-    s += series_to_pdf(res)
-    s += '>>'
-    to_pdf.update_object(catalogxref, s)
-
-
-
-def write_basic_ua(to_pdf, language=None):
-    # Create StructTreeRoot and descendants, allocate new xrefs as needed
-    structtreeroot_xref = to_pdf.get_new_xref()
-    parenttree_xref = to_pdf.get_new_xref()
-    page_info_xrefs = []
-    page_info_a_xrefs = []
-    parenttree_kids_xrefs = []
-    parenttree_kids_indirect_xrefs = []
-
-    kids_cnt = ceil(to_pdf.pageCount / 32)
-    for _ in range(kids_cnt):
-        kid_xref = to_pdf.get_new_xref()
-        parenttree_kids_xrefs.append(kid_xref)
-
-    # Parent tree contains a /Kids entry with a list of xrefs, that each contain
-    # a list of xrefs (limited to 32 per), and each entry in that list of list
-    # of xrefs contains a single reference that points to the page info xref.
-    for idx, page in enumerate(to_pdf):
-        page_info_xref = to_pdf.get_new_xref()
-        page_info_xrefs.append(page_info_xref)
-
-        page_info_a_xref = to_pdf.get_new_xref()
-        page_info_a_xrefs.append(page_info_a_xref)
-
-        parenttree_kids_indirect_xref = to_pdf.get_new_xref()
-        parenttree_kids_indirect_xrefs.append(parenttree_kids_indirect_xref)
-
-
-    for idx in range(kids_cnt):
-        start = idx*32
-        stop = (idx+1)*31
-        if stop > to_pdf.page_count:
-            stop = to_pdf.page_count- 1
-
-        s = """<<
-  /Limits [ %d %d ]
-""" % (start, stop - 1)
-        s += '  /Nums [ '
-
-        for pidx in range(start, stop):
-            s += '%d %d 0 R ' % (pidx, parenttree_kids_indirect_xrefs[pidx])
-
-            if idx % 7 == 0:
-                s = s[:-1] + '\n' + '      '
-
-        s += ']\n>>'
-
-        to_pdf.update_object(parenttree_kids_xrefs[idx], s)
-
-
-    for idx, page in enumerate(to_pdf):
-        intrect = tuple([int(x) for x in page.rect])
-
-        s = """<<
-  /BBox [ %d %d %d %d ]
-  /InlineAlign /Center
-  /O /Layout
-  /Placement /Block
->>
-""" % intrect
-        to_pdf.update_object(page_info_a_xrefs[idx], s)
-
-        s = """ <<
-  /A %d 0 R
-  /K 0
-  /P %d 0 R
-  /Pg %d 0 R
-  /S /Figure
->>""" % (page_info_a_xrefs[idx], structtreeroot_xref, page.xref)
-
-        to_pdf.update_object(page_info_xrefs[idx], s)
-
-
-    for idx, page in enumerate(to_pdf):
-        s = '[ %d 0 R ]' % page_info_a_xrefs[idx]
-        to_pdf.update_object(parenttree_kids_indirect_xrefs[idx], s)
-
-
-    K = '  /Kids [ '
-    for idx in range(kids_cnt):
-        K += '%d 0 R ' % parenttree_kids_xrefs[idx]
-
-        if idx % 7 == 0:
-            K = K[:-1] + '\n' + '      '
-
-    K += ']'
-    s = """<<
-%s
->>
-""" % K
-
-    to_pdf.update_object(parenttree_xref, s)
-
-    K = '  /K [ '
-    for idx, xref in enumerate(page_info_xrefs):
-        K += '%d 0 R ' % xref
-
-        if idx % 7 == 0:
-            K = K[:-1] + '\n' + '      '
-
-    K += ']'
-
-    to_pdf.update_object(structtreeroot_xref, """
-<<
-""" + K + """
-  /Type /StructTreeRoot
-  /ParentTree %d 0 R
->>
-""" % parenttree_xref)
-
-    #  TODO? /ClassMap 1006 0 R
-    #  TODO? /ParentTreeNextKey 198
-
-
-    # Update pages, add back xrefs
-    for idx, page in enumerate(to_pdf):
-        page_data = to_pdf.xref_object(page.xref)
-        page_data = page_data[:-2]
-
-        page_data += """
-  /StructParents %d
-""" % idx
-
-        page_data += """
-  /CropBox [ 0 0 %.1f %.1f ]
-""" % (page.rect[2], page.rect[3])
-
-        page_data += """
-  /Rotate 0
-"""
-        page_data += """
-  /Tabs /S
-"""
-        page_data += '>>'
-        to_pdf.update_object(page.xref, page_data)
-
-    catalogxref = to_pdf.pdf_catalog()
-    s = to_pdf.xref_object(to_pdf.pdf_catalog())
-    s = s[:-2]
-    s += """
-  /ViewerPreferences <<
-    /FitWindow true
-    /DisplayDocTitle true
-  >>
-"""
-    if language:
-        s += """
-  /Lang (%s)
-""" % language
-
-    s += """
-  /MarkInfo <<
-    /Marked true
-  >>
-"""
-    s += """
-  /StructTreeRoot %d 0 R
-""" % structtreeroot_xref
-
-    s += '>>'
-    to_pdf.update_object(catalogxref, s)
-
-
-
-def write_metadata(from_pdf, to_pdf, extra_metadata):
-    doc_md = from_pdf.metadata if from_pdf is not None else {}
-
-    doc_md['producer'] = PRODUCER
-
-    if 'url' in extra_metadata:
-        doc_md['keywords'] = extra_metadata['url']
-    if 'title' in extra_metadata:
-        doc_md['title'] = extra_metadata['title']
-    if 'author' in extra_metadata:
-        doc_md['author'] = extra_metadata['author']
-    if 'creator' in extra_metadata:
-        doc_md['creator'] = extra_metadata['creator']
-    if 'subject' in extra_metadata:
-        doc_md['subject'] = extra_metadata['subject']
-
-    current_time = 'D:' + datetime.utcnow().strftime('%Y%m%d%H%M%SZ')
-    if from_pdf is not None:
-        doc_md['creationDate'] = from_pdf.metadata['creationDate']
-    else:
-        doc_md['creationDate'] = current_time
-    doc_md['modDate'] = current_time
-
-    # Set PDF basic metadata
-    to_pdf.set_metadata(doc_md)
-
-    have_xmlmeta = (from_pdf is not None) and (from_pdf.xref_xml_metadata() > 0)
-    if have_xmlmeta:
-        xml_xref = from_pdf.xref_xml_metadata()
-
-        # Just copy the existing XML, perform no validity checks
-        xml_bytes = from_pdf.xrefStream(xml_xref)
-        to_pdf.set_xml_metadata(xml_bytes.decode('utf-8'))
-    else:
-        current_time = datetime.utcnow().isoformat(timespec='seconds') + 'Z'
-
-        stream='''<?xpacket begin="" id="W5M0MpCehiHzreSzNTczkc9d"?>
-        <x:xmpmeta xmlns:x="adobe:ns:meta/">
-          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
-            <rdf:Description rdf:about="" xmlns:xmp="http://ns.adobe.com/xap/1.0/">
-              <xmp:CreateDate>{createdate}</xmp:CreateDate>
-              <xmp:MetadataDate>{metadatadate}</xmp:MetadataDate>
-              <xmp:ModifyDate>{modifydate}</xmp:ModifyDate>
-              <xmp:CreatorTool>{creatortool}</xmp:CreatorTool>'''.format(creatortool=xmlescape(extra_metadata.get('creatortool', PRODUCER)),
-           createdate=current_time, metadatadate=current_time,
-           modifydate=current_time)
-
-        stream += '''
-            </rdf:Description>
-            <rdf:Description rdf:about="" xmlns:pdf="http://ns.adobe.com/pdf/1.3/">'''
-
-        if 'url' in extra_metadata:
-            stream += '''
-              <pdf:Keywords>{keywords}</pdf:Keywords>'''.format(keywords=xmlescape(extra_metadata['url']))
-
-        stream += '''
-              <pdf:Producer>{producer}</pdf:Producer>'''.format(producer=xmlescape(PRODUCER))
-
-        stream += '''
-            </rdf:Description>
-            <rdf:Description rdf:about="" xmlns:dc="http://purl.org/dc/elements/1.1/">'''
-
-        if extra_metadata.get('title'):
-            stream += '''
-              <dc:title>
-                <rdf:Alt>
-                  <rdf:li xml:lang="x-default">{title}</rdf:li>
-                </rdf:Alt>
-              </dc:title>'''.format(title=xmlescape(extra_metadata.get('title')))
-
-        # "An entity responsible for making the resource."
-        # https://www.dublincore.org/specifications/dublin-core/dcmi-terms/#http://purl.org/dc/terms/creator
-        # So should be author...
-        if extra_metadata.get('author'):
-            stream += '''
-              <dc:creator>
-                <rdf:Seq>
-                  <rdf:li>{author}</rdf:li>
-                </rdf:Seq>
-              </dc:creator>'''.format(author=xmlescape(extra_metadata.get('author')))
-
-        # TODO: Support multiple languages here?
-
-        if extra_metadata.get('language'):
-        # Empty language field means unknown language
-            stream += '''
-              <dc:language>
-                <rdf:Bag>'''
-
-            for language in extra_metadata.get('language', []):
-                stream += '''
-                  <rdf:li>{language}</rdf:li>'''.format(language=xmlescape(language))
-
-            stream += '''
-                </rdf:Bag>
-              </dc:language>'''
-
-        stream += '''
-            </rdf:Description>
-            <rdf:Description rdf:about="" xmlns:pdfaid="http://www.aiim.org/pdfa/ns/id/">
-              <pdfaid:part>3</pdfaid:part>
-              <pdfaid:conformance>B</pdfaid:conformance>
-            </rdf:Description>
-          </rdf:RDF>
-        </x:xmpmeta>
-        <?xpacket end="r"?>'''
-
-        to_pdf.set_xml_metadata(stream)
-
-
 # TODO: Document these options (like in bin/recode_pdf)
 def recode(from_pdf=None, from_imagestack=None, dpi=None, hocr_file=None,
         scandata_file=None, out_pdf=None, out_dir=None,
@@ -964,6 +640,7 @@ def recode(from_pdf=None, from_imagestack=None, dpi=None, hocr_file=None,
         mrc_image_format=None,
         downsample=None,
         bg_downsample=None,
+        fg_downsample=None,
         denoise_mask=None,
         hq_pages=None,
         hq_bg_compression_flags=None, hq_fg_compression_flags=None,
@@ -1078,6 +755,7 @@ def recode(from_pdf=None, from_imagestack=None, dpi=None, hocr_file=None,
                           jbig2=jbig2,
                           downsample=downsample,
                           bg_downsample=bg_downsample,
+                          fg_downsample=fg_downsample,
                           denoise_mask=denoise_mask,
                           reporter=reporter,
                           hq_pages=HQ_PAGES,
@@ -1171,6 +849,7 @@ def recode(from_pdf=None, from_imagestack=None, dpi=None, hocr_file=None,
         print('Compression ratio: %f' % (compression_ratio))
 
     # 5. Remove leftover files
+    outdoc.close()
     remove(tess_tmp_path)
 
     return {'errors': errors,
